;-------------------------------------------------------------------------------
;	iso7816_vcc.asm Version 1.0, January 13, 2005
;	Copyright (c) 2003-2005 by Matt Brunk
;	web:    http://www.smartcache.net/
;	e-mail: info@smartcache.net
;
;	The contents of this file are subject to the Artistic License (the "License").
;	You may not use this file except in compliance with the License. 
;	You may obtain a copy of the License at:
;	http://www.opensource.org/licenses/artistic-license.html
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;	This source code implements an interface between an RS-232 serial port and 
;	an ISO 7816 smart card, and is targeted for the Atmel AT90S2313 
;	microcontroller.  The frequency of the clock signal applied to the Atmel 
;	microcontroller is assumed to be 9.216 MHz.  As a result, the 10 MHz 
;	version of the microcontroller must be used (e.g., AT90S2313-10PC).  Do 
;	not use the 4 MHz version of the microcontroller (e.g., AT90S2313-4PC)--if 
;	you do, you will be exceeding the absolute maximum rating by more than a 
;	factor of 2.  Atmel has this to say in their datasheet:
;
;	NOTICE:	Stresses beyond those listed under "Absolute Maximum Ratings"
;			may cause permanent damage to the device. This is a stress
;			rating only and functional operation of the device at these or
;			other conditions beyond those indicated in the operational 
;			sections of this specification is not implied. Exposure to 
;			absolute maximum rating conditions for extended periods may 
;			affect device reliability. 
;
;	Data is transmitted over the RS-232 serial port at a baud rate of 115,200 
;	bps.  The frequency of the clock signal applied to the ISO 7816 CLK contact
;	is assumed to be 4.608 MHz.  See the pin assignments below for other 
;	assumptions.
;
;   Timer/Counter1 (Timer1) is used in Pulse Width Modulation (PWM) mode. The
;	PWM output on pin PB3 (OC1) is input into an RC low-pass filter, then into
;	an operational amplifier configured as a voltage follower (unity gain). The
;	output of the operational amplifier is applied to the ISO VCC and CLK
;	contacts via a 4053 analog multiplexer/demultiplexer.
;
;	Timer/Counter0 is used for bit timings of data sent to and received from
;	the smart card. It is also used to signal the end of data transfers. The
;	prescale value for Timer0 determines which mode it is being used in.
; 
;-------------------------------------------------------------------------------

.include "2313def.inc"

;-------------------------------------------------------------------------------
; AVR clock frequency is 9.216 MHz --> period is 108.5 ns
; ISO clock frequency is 4.608 MHz --> period is 217.0 ns
; ISO baud rate is ISO clock / 372, or 12387 bps
; ISO bit width is 80.73 us
; ISO data is start bit plus 8 data bits plus (even) parity bit plus
; two stop bits, 12 bits total for 969 us
; ISO RST to start bit is 520 us
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
; TIMER OVERFLOW TIMES for 9.216 MHz clock frequency
; 
; prescale  Timer0     Timer1
; --------  ---------  ---------
; CK         27.78 us    7.11 ms
; CK/8      222.20 us   56.90 ms
; CK/64       1.78 ms  455.10 ms
; CK/256      7.11 ms    1.82 s
; CK/1024    28.44 ms    7.30 s
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
; GLOBAL REGISTERS
;-------------------------------------------------------------------------------
.def	u_sr			= r1	; used to save SREG in ISRs
.def	u_buffer		= r14	; serial port data byte

.def	u_tmp			= r16	; scratch register
.def	u_bit_cnt		= r17	; number of bits transmitted or received for
								; current byte of data
.def	u0				= r18	; always all bits 0
.def	u_state			= r19	; bit-mask state
.def	u_xmit_char		= r20	; byte to transmit to PC or card
.def	A				= r21	; scratch register
.def	B				= r22	; available for debug
.def	C				= r23	; available for debug
.def	u_t0_counter	= r24	; count of Timer0 overflows
.def	parity_bit		= r25	; parity bit for serial data

;-------------------------------------------------------------------------------
;	Bit positions in u_state
;-------------------------------------------------------------------------------
.equ	RD=0					; receiving data from card
.equ	TD=1					; transmitting data to card

;-------------------------------------------------------------------------------
;	Timer0 constants for bit spacings.
;-------------------------------------------------------------------------------
.equ	BS1		= 0xA4			; one bit spacing
.equ	BS1_5	= 0x79			; 1.5 bit spacings

;-------------------------------------------------------------------------------
;	Reset and Interrupt Vectors
;-------------------------------------------------------------------------------
.cseg

.org 	$0000
rjmp	RESET_HANDLER			; Reset handler
.org	ICP1addr
rjmp	INT_ICP					; Input Capture interrupt handler
.org	OVF0addr
rjmp	INT_OVF_TIMER0			; TIMER0 overflow interrupt handler
.org 	URXCaddr
rjmp	INT_RX_CHAR 			; UART receive char interrupt handler

;-------------------------------------------------------------------------------
;	PIN ASSIGNMENTS
;-------------------------------------------------------------------------------
; ******
; PORT B
; ******
;
;   IN: SCK - clock for programming device and ISO RESET -- from RS-232 RTS
;   |   OUT: MISO - to RS-232 CTS
;   |   |   IN: MOSI - from RS-232 RTS inverted
;   |   |   |   IN: from ISO card slot switch; high means switch open and card in
;   |   |   |   |   OUT: OC1 - input to voltage follower for VCC
;   |   |   |   |   |   OUT: Used for bit timing debug
;   |   |   |   |   |   |   OUT: AIN1 - LEDs (write cathode, read anode)
;   |   |   |   |   |   |   |   OUT : AIN0 - LEDs (write anode, read cathode)
;   |   |   |   |   |   |   |   |
; +---+---+---+---+---+---+---+---+
; | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
; +---+---+---+---+---+---+---+---+
;  19  18  17  16  15  14  13  12
;
; ******
; PORT D
; ******
;
;   N/A
;   |   IN/OUT: ICP - I/O on ISO card slot
;   |   |   OUT: T1 - RESET on ISO card slot
;   |   |   |   OUT: T0 - 4053 MUX enable -- low to enable
;   |   |   |   |   OUT: INT1 - alternate clock enable
;   |   |   |   |   |   OUT: INT0 - ISO VCC Mux select
;   |   |   |   |   |   |   OUT: TXD - RS-232 transmit data
;   |   |   |   |   |   |   |   IN : RXD - RS-232 receive data 
;   |   |   |   |   |   |   |   |
; +---+---+---+---+---+---+---+---+
; | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
; +---+---+---+---+---+---+---+---+
;      11   9   8   7   6   3   2
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
; ISO 7816 contacts
;
;       -----------------------------------------------------------------
;      |      :  :
;      |      :  C
;      |      D  :
;      |      :  ---- ,----,            ,----,
;      |      :       | C1 |            | C5 | 
;      |      ------- '----'            '----' 
;      |              ,----,            ,----, 
;      |              | C2 |            | C6 | 
;      |              '----'            '----' 
;      |              ,----,            ,----, 
;      |              | C3 |            | C7 | 
;      |              '----'            '----' 
;      |              ,----,            ,----, 
;      |              | C4 |            | C8 | 
;      |              '----'            '----'
;      |                                :    :
;      |              A                 :    :
;      |<------------------------------>:    :
;      |                                     :
;      |                  B                  :
;      |<----------------------------------->:
;      |
;
;                        Assignment of the contacts
;   +---------------------------------------------------------------------+
;   | Contact No.|                     Assignment                         |
;   |------------+--------------------------------------------------------|
;   |    C1      | VCC (Supply voltage)                                   |
;   |------------+--------------------------------------------------------|
;   |    C2      | RST (Reset signal)                                     |
;   |------------+--------------------------------------------------------|
;   |    C3      | CLK (Clock signal)                                     |
;   |------------+--------------------------------------------------------|
;   |    C4      | Reserved for future use in other parts of ISO/IEC 7816 |
;   |------------+--------------------------------------------------------|
;   |    C5      | GND (Ground)                                           |
;   |------------+--------------------------------------------------------|
;   |    C6      | VPP Variable supply voltage (e.g. programming voltage) |
;   |------------+--------------------------------------------------------|
;   |    C7      | I/O (Data input/output)                                |
;   |------------+--------------------------------------------------------|
;   |    C8      | Reserved for future use in other parts of ISO/IEC 7816 |
;   +---------------------------------------------------------------------+
;
;   Location of the contacts:
;   ------------------------
;
;   Dimensions in millimeters
;
;       |   A       B       C       D    
;   ----+------------------------------- 
;    C1 | 10.25   12.25   19.23   20.93  
;    C2 | 10.25   12.25   21.77   23.47  
;    C3 | 10.25   12.25   24.31   26.01  
;    C4 | 10.25   12.25   26.85   28.55  
;    C5 | 17.87   19.87   19.23   20.93  
;    C6 | 17.87   19.87   21.77   23.47  
;    C7 | 17.87   19.87   24.31   26.01  
;    C8 | 17.87   19.87   28.85   28.55  
;   ----+------------------------------- 
;
;-------------------------------------------------------------------------------


;-------------------------------
; Start of code on device reset.
;-------------------------------
RESET_HANDLER:

	;---------------------------------------------------------------------------
	;	Initialize the stack pointer.
	;---------------------------------------------------------------------------
	ldi		A,RAMEND		
	out		SPL,A

	;---------------------------------------------------------------------------
    ;   u0 is initialized to all bits 0 and never changed.  It is used for 
	;	setting registers and output pins to 0. Registers B and C can be used 
	;	for debug.
	;---------------------------------------------------------------------------
	clr		u0			; zeroed and never changed
	clr		B			; can be used for debug
	clr		C			; can be used for debug

	;---------------------------------------------------------------------------
	;	Disable external interrupts.
	;---------------------------------------------------------------------------
	out		GIMSK,u0	

	;---------------------------------------------------------------------------
	;	Initialize port B.
	;---------------------------------------------------------------------------
	ldi		A,0xDB		; bit 7 from RS-232 RTS (set to 1 to enable pull-up)
	out		PORTB,A		; bit 6 to RS-232 CTS
						; bit 4 from ISO card switch (set to 1 to enable pull-up)
						; bit 3 PWM to RC low-pass filter and OP-AMP
						; bits 1 and 0 turn off LEDs

	ldi		A,0x4F		; set PORT B directions
	out		DDRB,A

	;---------------------------------------------------------------------------
	;	Initialize port D.
	;---------------------------------------------------------------------------
	ldi		A,0x62		; bit 6 to ISO I/O
	out		PORTD,A		; bit 5 to ISO RST--disable
						; bit 4 to 4053 MUX (active low)--enable
						; bit 3 use 4.608 MHz ISO clock
						; bit 2 to ISO VCC Mux--use output from PWM

	ldi		A,0x7E		; set PORT D directions
	out		DDRD,A
		
	;---------------------------------------------------------------------------
	;	Load PWM value from EEPROM.
	;---------------------------------------------------------------------------
	ldi		A,EE_PWM	; load the EEPROM address of the TOP value for PWM
	out		EEAR,A		; write the value to the EEPROM address register
	sbi		EECR,EERE	; send the Read strobe
	in		A,EEDR		; read the EEPROM data

	;---------------------------------------------------------------------------
	;	Set the PWM value to the value just loaded from EEPROM.  Clear the 
	;	high byte, as we are only using 8-bit PWM.  Note that the PWM value is 
	;	set only here and never changes.  Consequently, the nominal voltage 
	;	applied to the ISO VCC contact never changes.
	;---------------------------------------------------------------------------
	out		OCR1AH,u0
	out		OCR1AL,A

	;---------------------------------------------------------------------------
	;	Set up to use 8-bit, non-inverted PWM.  The PWM output is input into 
	;	an external RC low-pass filter and then into an OP-AMP configured as a 
	;	voltage follower.  The output of the OP-AMP is applied to the VCC and 
	;	CLK contacts of the smart card via the 4053 multiplexer.
	;---------------------------------------------------------------------------
	ldi		A,(1<<COM1A1)+(1<<PWM10)
	out		TCCR1A,A		; non-inverted and 8-bit PWM
	ldi		A,(1<<CS10)
	out		TCCR1B,A		; count at CK

	;---------------------------------------------------------------------------
	;	Initialize the UART (serial data interface to the PC).
	;---------------------------------------------------------------------------
	rcall	UART_INIT

	;---------------------------------------------------------------------------
	;	Enable interrupts.
	;---------------------------------------------------------------------------
	sei

	;---------------------------------------------------------------------------
	;	Wait for a card to be inserted.
	;---------------------------------------------------------------------------
wait_for_card_in:

	;---------------------------------------------------------------------------
	;	Read the state of the ISO card slot switch.  Card is inserted when 
	;	this pin is HIGH.
	;---------------------------------------------------------------------------
	sbis    PINB,4           ; skip if bit in I/O register set
	rjmp	wait_for_card_in

	;---------------------------------------------------------------------------
	;	Card is inserted.  Wait for start of RESET signal (RS-232 RTS line) 
	;	from serial port.
	;---------------------------------------------------------------------------
wait_for_reset_start:
	
	;---------------------------------------------------------------------------
	;	RESET starts when this pin is LOW
	;---------------------------------------------------------------------------
	sbic    PINB,7          ; skip if bit in I/O register cleared
	rjmp	wait_for_reset_start

	;---------------------------------------------------------------------------
	;	RESET is asserted.  Apply LOW to ISO RST contact to reset the card.
	;---------------------------------------------------------------------------
	cbi		PORTD,5

	;---------------------------------------------------------------------------
	;	Wait for RS-232 RTS line to indicate end of RESET.
	;---------------------------------------------------------------------------
wait_for_reset_end:

	;---------------------------------------------------------------------------
	;	RESET ends when this pin is HIGH
	;---------------------------------------------------------------------------
	sbis    PINB,7
	rjmp	wait_for_reset_end

	;---------------------------------------------------------------------------
	;	Apply HIGH to ISO RST contact to reset the card.
	;---------------------------------------------------------------------------
	sbi		PORTD,5
	
	cbi		PORTB,2			; initialize debug output on spare pin	

	;---------------------------------------------------------------------------
	;	Set up for the start bit (from first byte of ATR) from card.
	;---------------------------------------------------------------------------
	rcall	SETUP_FOR_START_BIT

	rjmp	wait_for_card_in

; end of AVR reset handler

;-------------------------------------------------------------------------------
;-------------------------------------------------------------------------------
;	SUBROUTINES
;-------------------------------------------------------------------------------
;-------------------------------------------------------------------------------
TURN_ON_WRITE_LED:
	sbi		PORTB,0		; write LED VCC
	cbi		PORTB,1		; write LED ground
	ret	

TURN_ON_READ_LED:
	cbi		PORTB,0		; read LED ground
	sbi		PORTB,1		; read LED VCC
	ret	

TURN_OFF_LEDS:
	;---------------------------------------------------------------------------
	;	Applying VCC to both anode and cathode of LEDs turns them off.
	;---------------------------------------------------------------------------
	sbi		PORTB,0		; set to VCC
	sbi		PORTB,1		; set to VCC
	ret

UART_INIT:
	ldi		A,4			; 115200 baud for 9.216 MHz clock
	out		UBRR,A
	ldi		A,0x9C		; rcv & xmit & rcv interrupt & 9-bit chars
	out		UCR,A		; 9th bit is used for even parity
	ret

CLEAR_TIMER_INTERRUPTS:
	;---------------------------------------------------------------------------
	;	Clear timer/counter interrupts.  These are cleared by writing a 
	;	logical "1" to the flags in the Timer/Counter Interrupt FLAG Register 
	;	(TIFR).
	;---------------------------------------------------------------------------
	ldi		u_tmp,(1<<TOV1)+(1<<ICF1)+(1<<TOV0)	
	out		TIFR,u_tmp
	ret

;-------------------------------------------------------------------------------
;	This sends a character over the UART (to the PC).  The 8-bit data to send 
;	must be in u_xmit_char and the parity bit in the LSB of parity_bit.
;-------------------------------------------------------------------------------
UART_SEND:

	;---------------------------------------------------------------------------
	;	The parity bit is transmitted as the 9th data bit (TXB8) in the UCR 
	;	register.
	;---------------------------------------------------------------------------
	sbrc	parity_bit,0		; skip the next instruction if the parity bit 
								; (bit 0 in parity_bit) is a 0
	sbi		UCR,TXB8			; the parity bit is a 1, so set TXB8 to a 1 in 
								; the UCR register
	sbrs	parity_bit,0		; skip the next instruction if the parity bit 
								; (bit 0 in parity_bit) is a 1
	cbi		UCR,TXB8			; the parity bit is a 0, so set TXB8 to a 0 in 
								; the UCR register

	out		UDR,u_xmit_char 	; transmit character over serial port
	ret

;-------------------------------------------------------------------------------
;	This sets up for the input capture interrupt to fire on the leading edge 
;	of the start bit from the card.
;-------------------------------------------------------------------------------
SETUP_FOR_START_BIT:

	;---------------------------------------------------------------------------
	;	Disable timer interrupts.
	;---------------------------------------------------------------------------
	out		TIMSK,u0

	;---------------------------------------------------------------------------
	;	Set ISO I/O line to input.
	;---------------------------------------------------------------------------
	sbi		PORTD,6	; set to enable pull-up
	cbi     DDRD,6
	cbi		PORTD,6	; set to tri-state

	;---------------------------------------------------------------------------
	;	Set our state to read data.
	;---------------------------------------------------------------------------
	ldi		u_state,1<<RD

	;---------------------------------------------------------------------------
	;	Clear timer/counter interrupts.
	;---------------------------------------------------------------------------
	rcall	CLEAR_TIMER_INTERRUPTS

	;---------------------------------------------------------------------------
	;	Zero Timer0 counter.
	;---------------------------------------------------------------------------
	out		TCNT0,u0

	;---------------------------------------------------------------------------
	;	Set clock prescale for Timer0.
	;---------------------------------------------------------------------------
	ldi		u_tmp,5				; CK/1024 -- at 9.216 MHXZ this gives 28 ms
	out		TCCR0,u_tmp
	ldi		u_t0_counter,12		; # of Timer0 interrupts before disabling Timer0.

	;---------------------------------------------------------------------------
	;	Set up the Input Capture to look for a falling slope, which will 
	;	represent the leading edge of the start bit.
	;---------------------------------------------------------------------------
	clr		u_bit_cnt			; clear bit counter
	ldi		u_tmp,(1<<TOIE0)+(1<<TICIE)
	out		TIMSK,u_tmp			; interrupt at Timer0 and input capture

	ret

;-------------------------------------------------------------------------------
;	This performs Timer0 overflow processing when running at the coarse 
;	prescale (i.e., not at the finer prescale used for data bit timing).  
;	Timer0 is used to turn off the read LED and clear all timer interrupts.  
;	Because Timer0 is only an 8-bit value, its overflow period is too short 
;	for the duration needed to insure that all data transfers have completed.  
;	Therefore, we use a counter (u_t0_counter) that gets decremented by 1 for 
;	each Timer0 overflow.  When the counter is 0, the entire wait period ends.
;-------------------------------------------------------------------------------
TIMER0_OVF_COARSE:
	dec		u_t0_counter
	breq	timer0_ovf_coarse_clear	; branch if u_t0_counter is 0
	out		TCNT0,u0				; zero timer0 counter
	ret

timer0_ovf_coarse_clear:
	;---------------------------------------------------------------------------
	;	Turn off LEDs.
	;---------------------------------------------------------------------------
	rcall	TURN_OFF_LEDS

	;---------------------------------------------------------------------------
	;	Stop Timer0.
	;---------------------------------------------------------------------------
	out		TCCR0,u0

	;---------------------------------------------------------------------------
	;	Disable all timer interrupt flags.
	;---------------------------------------------------------------------------
	out		TIMSK,u0

	;---------------------------------------------------------------------------
	;	Clear timer/counter interrupts.
	;---------------------------------------------------------------------------
	rcall	CLEAR_TIMER_INTERRUPTS
	ret

;-------------------------------------------------------------------------------
;	UART receive character interrupt vector.  This is called when a character 
;	is received from the PC over the serial port.
;-------------------------------------------------------------------------------
INT_RX_CHAR:
	in		u_sr,SREG			; save the status register

	in		u_buffer,UDR 		; read serial data

	;---------------------------------------------------------------------------
	;	Save the parity bit.
	;---------------------------------------------------------------------------
	ldi		parity_bit,1
	sbis	UCR,RXB8
	ldi		parity_bit,0

	;---------------------------------------------------------------------------
	;	Set ISO I/O line direction to output.  We will be transmitting the 
	;	received byte to the card.
	;---------------------------------------------------------------------------
	sbi		PORTD,6
	sbi     DDRD,6

	;---------------------------------------------------------------------------
	;	Clear ISO I/O line (this represents the leading edge of the start bit).
	;---------------------------------------------------------------------------
	cbi		PORTD,6


	;---------------------------------------------------------------------------
	;	Set the current value of Timer0 so that it overflows 1 bit spacing 
	;	from now (which is when the start bit ends and the first data bit 
	;	begins).
	;---------------------------------------------------------------------------
	ldi		u_tmp,BS1
	out		TCNT0,u_tmp

	;---------------------------------------------------------------------------
	;	Use Timer0 for timing data bit spacings.  Note that the Timer0 
	;	prescale is changed to CK/8 when using Timer0 for this purpose.
	;---------------------------------------------------------------------------
	ldi		u_tmp,2
	out		TCCR0,u_tmp			; count at CK/8 rate
	rcall	CLEAR_TIMER_INTERRUPTS
	ldi		u_tmp,(1<<TOIE0)
	out		TIMSK,u_tmp			; interrupt at TC01 only

	;---------------------------------------------------------------------------
	;	Set up our state.
	;---------------------------------------------------------------------------
	mov		u_xmit_char,u_buffer
	ldi		u_state,1<<TD		; set Transmit Data flag (clear all others)
	clr		u_bit_cnt

	;---------------------------------------------------------------------------
	;	Turn on the write LED.
	;---------------------------------------------------------------------------
	rcall	TURN_ON_WRITE_LED

	out		SREG,u_sr			; restore the status register
	reti						; return from ISR

;-------------------------------------------------------------------------------
;	Input capture interrupt vector.  This is called at the leading edge of the 
;	start bit (data sent from card).
;-------------------------------------------------------------------------------
INT_ICP:
	in		u_sr,SREG			; save the status register

	clr		u_buffer			; clear buffer
	clr		u_bit_cnt			; clear bit counter
	out		TCCR0,u0			; disable Timer0 counter

	;---------------------------------------------------------------------------
	;	Set up Timer0 clock divider.
	;---------------------------------------------------------------------------
	ldi		u_tmp,2				;CK/8 -- at 9.216 MHZ this gives 222 us
	out		TCCR0,u_tmp

	;---------------------------------------------------------------------------
	;	Set up Timer0 to go off 1.5 bit spacings from now.  This gets us to 
	;	the middle of the first data bit (1 bit spacing for start bit plus 1/2 
	;	bit spacing for data bit).
	;---------------------------------------------------------------------------
	rcall	CLEAR_TIMER_INTERRUPTS
	ldi		u_tmp,(1<<TOIE0)
	out		TIMSK,u_tmp			; Interrupt at Timer0 overflow only
	ldi		u_tmp,BS1_5			; 1.5 bit spacings
	out		TCNT0,u_tmp

	rcall	TURN_ON_READ_LED	; turn on read LED

	out		SREG,u_sr			; restore the status register
	reti						; return from ISR

;-------------------------------------------------------------------------------
;	Timer0 overflow interrupt vector.
;-------------------------------------------------------------------------------
INT_OVF_TIMER0:
	in		u_sr,SREG			; save the status register

	;---------------------------------------------------------------------------
	;	The prescale value in TCCR0 is used to differentiate what Timer0 is 
	;	being used for.
	;---------------------------------------------------------------------------
	in		u_tmp,TCCR0
	cpi		u_tmp,2
	breq	tim0_io_bits
	rcall	TIMER0_OVF_COARSE
	rjmp	tim0_ret			; jump to return

;-------------------------------------------------------------------------------
;	Timer0 is being used for data bit timing if we get here.
;-------------------------------------------------------------------------------
tim0_io_bits:

	;---------------------------------------------------------------------------
	;	Set up Timer0 to overflow 1 bit spacing later.
	;---------------------------------------------------------------------------
	ldi		u_tmp,BS1
	out		TCNT0,u_tmp

	;---------------------------------------------------------------------------
	;	Toggle the state of pin 14.  This allows for bit transmit and sample 
	;	times to be monitored with an oscilloscope.
	;---------------------------------------------------------------------------
	rcall	DEBUG_TOGGLE_PIN14

	;---------------------------------------------------------------------------
	;	Update our state info.
	;---------------------------------------------------------------------------
	inc		u_bit_cnt			; increment bit count
	sbrs	u_state,TD			; if transmit-bit set
	rjmp	tim0_receive		; goto receive

;-------------------------------------------------------------------------------
;	Transmits (to the card) are done here.
;-------------------------------------------------------------------------------
tim0_xmit:

	;---------------------------------------------------------------------------
	;	Check if this is the parity or stop bit.
	;---------------------------------------------------------------------------
	cpi		u_bit_cnt,9			; if this is the 9th bit
	breq	tim0_xmit_parity	; transmit the parity bit
	cpi		u_bit_cnt,10		; if this is the 10th bit
	breq	tim0_xmit_stop		; transmit the stop bit

	;---------------------------------------------------------------------------
	;	Transmit the next data bit.
	;---------------------------------------------------------------------------
	sbrc	u_buffer,0			; skip the next instruction if the next data bit 
								; (bit 0 in u_buffer) is a 0
	sbi		PORTD,6				; the next data bit is a 1, so transmit a 1
	sbrs	u_buffer,0			; skip the next instruction if the next data bit 
								; (bit 0 in u_buffer) is a 1
	cbi		PORTD,6				; the next data bit is a 0, so transmit a 0
	lsr		u_buffer			; shift buffer right
	
	rjmp	tim0_ret			; jump to return

;-------------------------------------------------------------------------------
;	Transmit of the parity bit (to the card) is done here.
;-------------------------------------------------------------------------------
tim0_xmit_parity:
	sbrc	parity_bit,0		; skip the next instruction if the parity bit 
								; (bit 0 in parity_bit) is a 0
	sbi		PORTD,6				; the parity bit is a 1, so transmit a 1
	sbrs	parity_bit,0		; skip the next instruction if the parity bit 
								; (bit 0 in parity_bit) is a 1
	cbi		PORTD,6				; the parity bit is a 0, so transmit a 0

	rjmp	tim0_ret			; jump to return


;-------------------------------------------------------------------------------
;	Transmit of the stop bit (to the card) is done here.
;-------------------------------------------------------------------------------
tim0_xmit_stop:
	sbi		PORTD,6				; stop bit is transmitted as a 1
	rjmp	tim0_xmit_complete	; jump to xmit complete

;-------------------------------------------------------------------------------
;	Transmit of all bits to the card is complete when we get here.  Now echo 
;	the character back to the serial port (PC) and set up for data to be sent 
;	by the card.
;-------------------------------------------------------------------------------
tim0_xmit_complete:
	
	rcall	UART_SEND
	rcall	SETUP_FOR_START_BIT	; set up for start bit from card
	rjmp	tim0_ret			; jump to return

tim0_rcv_complete:
	;---------------------------------------------------------------------------
	;	Stop Timer0 and setup for next start bit from card.
	;---------------------------------------------------------------------------
	rcall	SETUP_FOR_START_BIT
	rjmp	tim0_ret			; jump to return
	
tim0_receive:

	;---------------------------------------------------------------------------
	;	The 10th bit is the stop bit (8 data bit plus parity bit are first 9).
	;---------------------------------------------------------------------------
	cpi		u_bit_cnt,10
	breq	tim0_rcv_complete

	;---------------------------------------------------------------------------
	;	Sample the data bit from the card.  We use the carry bit to store that 
	;	bit value, then shift the carry bit into the data byte we are building.
	;---------------------------------------------------------------------------
	sec							; set carry
	sbis	PIND,6				; skip if ISO I/O line is high.
	clc							; clear carry
	ror		u_buffer			; shift carry into data and bit 0 into carry

	in		u_tmp,SREG			; save SREG for the carry bit

	;---------------------------------------------------------------------------
	;	Exit if this is not the parity bit.
	;---------------------------------------------------------------------------
	cpi		u_bit_cnt,9			; must sample parity bit
	brne	tim0_ret			; jump to return

	;---------------------------------------------------------------------------
	;	Save the parity bit, which is in bit 7 that was just read.
	;---------------------------------------------------------------------------
	ldi		parity_bit,1
	sbrs	u_buffer,7
	ldi		parity_bit,0

	out		SREG,u_tmp			; restore SREG that has the carry bit 
								; (bit 0 of data from card)

	;---------------------------------------------------------------------------
	;	At this point, the parity bit is in bit 7 of u_buffer, bits 7 through 
	;	1 of the data from the card are in bits 6 through 0 of u_buffer, and 
	;	bit 0 of the data from the card is in the carry bit.  The "rol" 
	;	instruction is "Rotate Left Through Carry", which will shift all bits 
	;	in u_buffer left by one and put the carry bit in bit 0 of u_buffer.  
	;	This gives us the data byte sent from to card.
	;---------------------------------------------------------------------------
	rol		u_buffer

	;---------------------------------------------------------------------------
	;	Now transmit the data byte received from the card to the serial port.
	;---------------------------------------------------------------------------
	mov		u_xmit_char,u_buffer
	rcall	UART_SEND
	rjmp	tim0_ret			; jump to return

tim0_ret:
	out		SREG,u_sr			; restore status register
	reti						; return from ISR

;-------------------------------------------------------------------------------
;	This toggles the state of pin 14 (PB2).
;-------------------------------------------------------------------------------
DEBUG_TOGGLE_PIN14:
	in		u_tmp,PORTB	; read the state of PORT B
	sbrc	u_tmp,2		; skip if bit is cleared
	cbi		PORTB,2		; bit is set, so clear it
	sbrs	u_tmp,2		; skip if bit is set
	sbi		PORTB,2		; bit is clear, so set it
	ret

;-------------------------------------------------------------------------------
;	EEPROM data
;-------------------------------------------------------------------------------
.eseg
.org 0
EE_00:
.db 00,00
;-------------------------------------------------------------------------------
;	The first byte of the following data specifies the 8-bit value (the TOP 
;	value) that defines the pulse width used in PWM.  The pulse width (and 
;	consequently the duty cycle) is directly proportional to this value.  The 
;	larger the value, the wider the pulse width (and the larger the duty 
;	cycle).  The value of VCC applied to the ISO VCC and CLK contacts is 
;	directly proportional to this value.
;-------------------------------------------------------------------------------
EE_PWM:
.db 0xFF,00

